c FORTRAN90 code: Powers of S_{n}^{a}    (Ling-Tong), 
c                           Q_{n}(m)     (Box-Ljung), and 
c                           Q^{2}_{n](m) (Li-Mak)
c
c                 Table 2 (significance level = 0.05) of Ling-Tong (2011)
c                 H0: ARMA(1,1) model simulation
c                 H1: BL model (CHANGE for other models)
c                 
c File: scoreARMA.f (coded by Shiqing Lin, and modified by JDG)
c
c Reference:
c Ling, S. and Tong, H. (2011)
c   Score based goodness-of-fit tests for time series.
c   Statistica Sinica, 21, 1807--1829
c
c IMSL (double precision) routines:
c   DBCOAH = minimize a function of N variables subject to
c            bounds the variables using a modified Newton method
c           and a user-supplied Hessian.
c   DSVRGN = sort a real array by algebraic value.
c   DMURRV = multiply a real rectangular matrix by a vector.
c   DLINRG = compute the inverse of a real general matrix.       
c****************************************************************
       USE imsl
       implicit real*8(a-h, o-z)
       parameter (nn=20000,kki=2000,kbs=400,nc=20000,m=6)
       parameter (n5=2,nx=5000)
       dimension al(nn),e(nn),u(nn),y(nn),x(nx)
       dimension y1(kbs),y2(kbs),q(kki),arh(kki)
       dimension h(nn),cv(nc),cv1(nc),vv(n5,n5),r(m)
       dimension ga(nn),dd(2,nx),RY(Nx),AT(N5)
       dimension d2X(N5),d22X(N5,N5),Td22X(N5,N5)      
       dimension rparma(7),xx(n5),xguess(n5),xlb(n5),xscale(n5)
       dimension xub(n5),g1(n5),gg1(n5,n5),iparma(7),dx(nn)
       dimension d22A(N5,N5), Td22a(n5,n5),Tgg1(n5,n5)
       common    hh,g1,gg1,x,knk
       external  grad,hess,fcn
c       
c      Sample size        
       knk  = 400
       aknk = float(knk) 
       
c      true value of d
c      d = 0.3
         
c      Generate a series of coefficients of binomial polynomial
c      nga = 250
c      do i=1,nga
c         ga(i) = 1
c         do k=1,i
c            ak = float(k)
c            ga(i) = ga(i)*(d+ak-1)/ak
c         enddo
c      enddo
c
c      True values of parameters
c      Critical value
c      do ikk=1, nc
c         iseed = 1466+ikk*7+7+2+20+1
c         call  rnset(iseed)
c         call  drnnor(nn, al)
c         cr  = 0.0
c         nck = 4000
c         anck = float(nck)
c         do i=1, nck
c            crt = 0.0
c            do j=1, i
c               crt = crt+al(J)/sqrt(anck)
c            enddo     
c            if (cr.le.abs(crt)) then
c              cr = abs(crt)
c            endif     
c         enddo                           
c         cv(ikk) = cr                            
c         write(6,*) ikk
c      enddo          

c      call dsvrgn(nc,cv,cv1)
c 
c      ir0 = int(nc*0.995)
c      ir1 = int(nc*0.99)
c      ir2 = int(nc*0.975)                   
c      ir3 = int(nc*0.95)	
c      ir4 = int(nc*0.90)
c      ir5 = int(nc*0.80)
c      write(6,122)  cv1(ir0)**2,cv1(ir1)**2,cv1(ir2)**2
c      write(6,122)  cv1(ir3)**2,cv1(ir4)**2,cv1(ir5)**2                   
C      n = 200    7.2432    4.4048    3.5178
c      nck = 4000 nc = 20000    8.8240    7.6343    6.0174
c                              4.9112    3.7882    2.6613  
c***********************************************************************
c      kki = Number of replications        
       do 99 ikk=1001,kki
            write(6,*)(ikk)
            iseed = 1466+ikk*7+7+2+20+1
            call rnset(iseed)
            call drnnor(nn, al)

            a  = 0.5
            a0 = a
            b  = 0.5
            b0 = b
            
            do i=3, nx
c              H1: BL Model with theta = 0.2    
               u(i)=a*u(i-1)+b*al(i-1)+al(i)-0.2*u(i-2)*al(i-1)
            enddo

            do i=1,knk
               x(i)  = u(i+nx-knk)
               y1(i) = u(i+nx-knk)                         
            enddo
                  
            call erset(0,0,0)
            
c           Starting values
c           data xguess/0.1,0.8/
            data xscale/1.0,1.0/
            data fscale/1.0/
            xguess(1) = a0  
            xguess(2) = b0  
c ************************************************
c           Bounds of parameter estimates         
            data xlb/-1.0,-1.0/
     +           xub/1.0, 1.0/
            iparma(1) = 0
c           iparma(3) = 550
            ipp=0
c           Estimation of parameters         
            call dbcoah(fcn,grad,hess,n5,xguess,ipp,xlb,xub,xscale,
     +                  fscale,iparma,rparma,xx,f)         
c           Input results         
c                              
            a    = xx(1)
            b    = xx(2)
            adm  = b+adm
            adm2 = b*b+adm2
            sa   = sa+a
            saa  = saa+a*a

C  TEST STATISTICS ****************************************************
           call estimat1(x,knk,xx,hh,g1,gg1,ry,dd) 
           call dsvrgn(kbs,y1,y2)
  	   ir0 = int(knk*0.05)
           do ii=1,2
              do ii2=1,2
                 d22a(ii,ii2) = 0.0
              enddo
           enddo

           do ii=1,2
              do ii2=1,2
                 do i=2,knk
                    if (x(i-1) .le. y2(ir0) ) then
           d22a(ii,ii2)=d22a(ii,ii2)+dd(ii,i)*dd(ii2,i)/aknk
                    endif
                  enddo
               enddo
           enddo
   	                    
           CALL dlinrg(N5,d22a,N5,Td22a,N5)
           CALL dlinrg(N5,gg1,N5,Tgg1,N5)

            aa = 0.0
            do ii=1,2
               do ii2=1,2
                  aa = aa+Td22a(ii,ii2)+Tgg1(ii,ii2)
               enddo
            enddo

            ATN = 0.0                         
            do i=ir0, knk	                    
               do ii=1,2
                  d2x(ii) = 0.0
                  do ii2=1,2
                     d22x(ii,ii2) = 0.0
                  enddo
                enddo

                do J=2,knk
                   if (x(J-1).le.y2(i)) then
                   do ii=1,2
                d2x(ii) = d2x(ii)+ry(J)*dd(ii,J)/sqrt(aknk)
                      do ii2=1,2
                d22x(ii,ii2) = d22x(ii,ii2)+dd(ii,J)*dd(ii2,J)/aknk
                      enddo
                    enddo
                    endif
                enddo

                CALL dlinrg(N5,d22x,N5,Td22x,N5)
                call dmurrv(N5,N5,Td22x,N5,n5,d2x,1,N5,AT)
                ax  = AT(1)+AT(2)                       
                ATX = AX*AX/HH                   
                IF (ATX.GE.ATN) THEN
                  ATN = ATX
                ENDIF                          

            ENDDO   
c
c      Critical values of standard Brownian motion 
c      taken from Shorack and Wellner (1986, p. 34)
c      b = 5.00   (10%)
c      b = 7.6343 (5%)   
c      b = 3.83   (1%)

       If (ATN/AA .ge.7.634) then
          at01 = at01+1
       endif   
       If (ATN/AA .ge.5.00) then
          at05 = at05+1
       endif   
       If (ATN/AA .ge.3.83 ) then
          at1 = at1+1
       endif   
       write(6,*) 'Ling-Tong test statistic'
       write(6,*)  at01, at05, at1
c                 
c      Box-Ljung testing                    
       ILL=6
       do iL=1,ILL
          q(IL)   = 0.0
          arh(iL) = 0.0
       enddo  

       do iL=1, ILL
          ail = float(iL)
          do i= iL+1, knk
             arh(iL) = arh(iL)+ry(i)*ry(i-iL)/((aknk-ail)*hh)
          enddo                      
       enddo                       

       q(1) = aknk*(aknk+2)*arh(1)*arh(1)/(aknk-1)
       do iL=2,ILL
          q(IL) = q(IL-1)+aknk*(aknk+2)*arh(iL)*arh(iL)/(aknk-aiL)
       enddo
                 
       If (q(6).ge.13.277) then ! critical value Chi^2_{0.01} 4df
         bl1=bl1+1
       endif   
       If (q(6).ge.9.488) then  ! critical value Chi^2_{0.05} 4df
         bl05=bl05+1
       endif   
       If (q(6).ge.7.779) then  ! critical value Chi^2_{0.1} 4df
         bl01=bl01+1
       endif   

       write(6,*)  'Box-Ljung test statistic'
       write(6,122)  Bl1, Bl05, Bl01                  

c      Li-Mak testing                    
       c0 = 0.0
       do i=1,knk
         c0 = c0+(((ry(i)**2)/hh-sc1)**2)/AKNK
       enddo

       sc1 = 0.0
       do i=3, knk
         sc1 = sc1+((ry(i)**2)/hh)/aknk
       enddo

       do i=1,m
         r(i) = 0
       enddo

       do i=1,m
         do k=i+1,knk
            r(i) = r(i)+((ry(k)**2)/hh-sc1)*((ry(k-i)**2)/hh
     +            -sc1)/(c0*(aknk))
         enddo
       enddo 

       do i=1,n5
         do ii=1,n5
            vv(i,ii) = 0.0
         enddo
         vv(i,i) = 1.0
       enddo

       qm=(Aknk-1)*dblinf(m,m,vv,m,r,r)
       IF (qm.GE.10.645) THEN ! critical value Chi^{2}_{0.1} 6 df
          RJECT10 = RJECT10+1
       ENDIF   
       IF (qm.GE.12.592) THEN ! critical value Chi^{2}_{0.05} 6 df
          RJECT05 = RJECT05+1
       ENDIF   
       IF (qm.GE.16.812) THEN ! critical value Chi^{2}_{0.01} 6 df
           RJECT01 = RJECT01+1
       ENDIF   
       
       write(6,*)  'Li-Mak test statistic'
       WRITE(6,122) RJECT01, RJECT05, RJECT10
c **************************************************
   99  continue  !  end loop ikk=1001,kki 
c   
c      print final results
       ak = float(kki)
       
       write(6,*)('Number of replications, Sample size')
       write(6,*) kki-1000, knk
       write(6,*)('Table parameters')
       write(6,120) a0, b0
  120  format(' ',4F10.4)
  
       admm  = adm/ak
       admm2 = adm2/ak-admm**2
       udmm2 = dsqrt(admm2)
       write(6,*)('Mean of b and empirical standard errors of b')
       write(6,122) admm,dsqrt(admm2)
       
       write(6,*)('Mean of a and empirical standard errors of a')
       write(6,122) sa/ak,dsqrt(saa/ak-(sa/ak)**2)
  122  format(' ',4F10.4)

       end
        
        subroutine fcn(n5,xx,f)
        implicit real*8(a-h,o-z)
        parameter (nx=5000)
        dimension xx(n5),g1(2),gg1(2,2),x(nx),ry(nx)
        dimension dd(2,nx)
        common    hh,g1,gg1,x,knk
        call estimat1(x,knk,xx,hh,g1,gg1,ry,dd)
        f=hh
        return
        end
        
        subroutine grad(n5,xx,g)
        implicit real*8(a-h,o-z)
        parameter (nx=5000)
        dimension xx(n5),g(n5),g1(2),gg1(2,2),x(nx)
        common    hh,g1,gg1,x,knk
        do i=1,n5
           g(i) = g1(i)
        enddo
        return
        end
                      
        subroutine hess(n5,xx,gg,ldh)
        implicit real*8(a-h,o-z)
        parameter(nx=5000)
        dimension xx(n5),gg(ldh,n5),g1(2),gg1(2,2),x(nx)
        common hh,g1,gg1,x,knk
        do i=1,n5
           do ii=1,n5
              gg(i,ii) = gg1(i,ii)
           enddo
        enddo
        return
        end
        
c       The following program calculates gradient and information matrix.
        
        subroutine estimat1(x,knk,xx,hh,g1,gg1,ry,dd)
        implicit real*8(a-h,o-z)
        parameter (nn=1000,n5=2)
        parameter (nx=5000)
        dimension x(nx),y(nn),ed(nn),d2(2),d22(2,2)
        dimension ry(nn),dd(2,nn)
        dimension ga(nn)
        dimension g1(n5),gg1(n5,n5),xx(n5)
        
        a = xx(1)
        b = xx(2)
        aknk  = float(knk)
        ry(1) = x(1)
        do i=2,knk
           ry(i) = x(i)-a*x(i-1)-b*ry(i-1)
        enddo
                              
        hh=0
        do i=2,knk
           hh = hh+ry(i)*ry(i)/(aknk-2.0)
        enddo

        do i=1,knk
           dd(1,i)=0.0
           dd(2,i)=0.0
        enddo

        do i=2,knk
           dd(1,i) = -x(i-1)-b*dd(1,i-1)
           dd(2,i) = -ry(i-1)-b*dd(2,i-1)
        enddo

        do i=1,2
           d2(i) = 0
           do ii=1,2
              d22(i,ii) = 0
           enddo
        enddo
        do ii=1,2
           do i=2,knk
              d2(ii) = d2(ii)+dd(ii,i)*ry(i)
           enddo
        enddo
        do ii=1,2
           do ii2=1,2
              do i=2,knk
                 d22(ii,ii2) = d22(ii,ii2)+dd(ii,i)*dd(ii2,i)
              enddo
           enddo
        enddo
        
        g1(1) = d2(1)/aknk
        g1(2) = d2(2)/aknk
        do i=1,2
           do ii=1,2
              gg1(i,ii) = -d22(i,ii)/aknk
           enddo
        enddo

        return
        end
