c FORTRAN90 code: Powers of Sn^{a}       (Ling-Tong), 
c                           Q_{n}(m)     (Box-Ljung), and 
c                           Q^{2}_{n](m) (Li-Mak)
c                                         
c                 Table 5 (significance level = 0.05) of Ling-Tong (2011)
c                 H0: GARCH(1,1) model simulation
c                     e(t)=al(t)* sqrt{h(t)); 
c                     h(t)=c(1)+c(2) e(t)*e(t)+c(3)h(t-1),  
c                     and al(t)=standard normal N(0,1) 
c                 H1: NAGARCH model with theta = 1.2 (CHANGE for other models)
c                 
c File: scoreGARCH.f (coded by Shiqing Lin, and modified by JDG)
c
c Reference:
c Ling, S. and Tong, H. (2011)
c   Score based goodness-of-fit tests for time series.
c   Statistica Sinica, 21, 1807--1829
c
c IMSL (double precision) routines:
c   DBCOAH = minimize a function of N variables subject to
c            bounds the variables using a modified Newton method
c            and a user-supplied Hessian.
c   DSVRGN = sort a real array by algebraic value.
c   DMURRV = multiply a real rectangular matrix by a vector.
c   DLINRG = compute the inverse of a real general matrix.       
c
c****************************************************************
       USE imsl
       
       implicit real*8(a-h, o-z)
       parameter (nn=8000,kki=1000,ir=3,KBS=400,m=6)
       parameter (n5=3,nx=8000)
       dimension al(nn),c(3),c0(3),h(nn),e(nn),y(nx)
       dimension rparma(7),xx(n5),xguess(n5),xlb(n5),xscale(n5)
       dimension xub(n5),g1(n5),gg1(n5,n5),iparma(7),dx(nn),hdd(ir,nx)
       dimension axx(n5),axx2(n5),h2(nx),ry(nx),ak(nx)
       dimension AS(nx),AST(nx),eta(nn),w(ir,nn)
       dimension dgg1(n5,n5),hd(1,nx),dd(ir,nx),q(m),arh(m)
       dimension AT(N5),Y2(KBS),Y1(kbs)
       dimension d2X(N5),d22X(N5,N5),Td22X(N5,N5)      
       dimension d22A(N5,N5),Td22A(N5,N5),TGG1(N5,N5)
       
       common    hh,g1,gg1,y,knk 
       external  grad,hess,fcn

       knk  = 400
       aknk = float(knk)
c   **************************************************
c      kki = Number of replications

       do 99 ikk=1,kki
             write(6,*)(ikk)
             iseed=1466+ikk*100+7+2+20
             call rnset(iseed)
             call drnnor(nn, al)

c            True values of parameters
c       
             c(2)  = 0.2  ! alpha1
             c(3)  = 0.8  ! beta1
             c(1)  = 0.01
             c0(1) = c(1)
             c0(2) = c(2)
             c0(3) = c(3)
             h(1)  = 1.0
             
	delta=1.5
        do i=3,nn          
           h(i) = (0.1+0.4*H(I-1)**(delta/2)+0.3*abs(e(I-1)
     +             -1.2*abs(al(i)))**delta)**(2.0/delta) ! H1: NAGARCH model
                                                         ! theta = 1.2
           hi=0.0
           e(i) = dsqrt(h(i))*al(i)    
        enddo

c       True observations
c 
        do i=1,knk
           y(i)  = e(i+nn-knk)
           y1(i) = y(i)
        enddo       
c       
        call erset(0,0,0)
        
c       Starting values     
c       data xguess/0.05,0.1,0.85/
        xguess(1) = c0(1)
        xguess(2) = c0(2)
        xguess(3) = c0(3)

c       xguess(1) = 0.1
c       xguess(2) = 0.2
c       xguess(3) = 0.2
c
        data xscale/1.0,1.0,1.0/
        data fscale/1.0/
c
c       Bounds of parameters estimated
c  *********************************************************
c
         data xlb/0.00005, 0.0,0.0/,
     +        xub/1.0,     1.0,1.0/
         iparma(1) = 0
         iparma(3) = 550
         ipp = 0
c        Estimation of parameters        
         call dbcoah(fcn,grad,hess,n5,xguess,ipp,xlb,xub,xscale,
     +               fscale,iparma,rparma,xx,f)

c  *******************************************************************
          akki = float(kki)
          do i=1,n5
             axx(i)  = axx(i)+xx(i)/akki
             axx2(i) = axx2(i)+(xx(i)**2)/akki
          enddo
           
C  TEST STATISTICS ******************************************************
         call estimat1(y,knk,xx,g1,gg1,h2,ry,hdd,hh,dgg1)
         call dsvrgn(kbs,y1,y2)
         ir0=int(knk*0.10)
	
         ah4 = 1.0
         do ii=1,n5
            do ii2=1,n5
               d22a(ii,ii2) = 0.0
            enddo
         enddo

         do ii=1,N5
            do ii2=1,N5
               do i=2,knk
                  if (Y(i-1) .le. y2(ir0) ) then
             d22a(ii,ii2)=d22a(ii,ii2)+hdd(ii,i)*hdd(ii2,i)
     +                /(2*(aknk-2)*h2(i)*h2(i))
                  endif
               enddo
            enddo
         enddo
   	
         CALL dlinrg(N5,d22a,N5,Td22a,N5)
         CALL dlinrg(N5,gg1,N5,Tgg1,N5)
         aa = 0.0
         do ii=1,N5
            do ii2=1,N5
               aa = aa+Td22a(ii,ii2)+Tgg1(ii,ii2)
            enddo
         enddo

         ATN = 0.0
 	 do i=ir0, knk
 	    do ii=1,N5
               d2x(ii) = 0.0
               do ii2=1,N5
                  d22x(ii,ii2) = 0.0
               enddo
             enddo

             do J=2,knk
                if (Y(J-1).le.y2(i)) then
                  do ii=1,n5
                      d2x(ii) = d2x(ii)- ( ry(J)*ry(J)/h2(J)-1 )
     +                   *hdd(ii,J)/(2*h2(J)*sqrt(aknk-2.0))
                     do ii2=1,n5
                      d22x(ii,ii2)=d22x(ii,ii2)+hdd(ii,J)*hdd(ii2,J)
     +                /(2*(aknk-2.0)*h2(J)*h2(J) )
                    enddo
                  enddo
                endif
              enddo

            CALL dlinrg(N5,d22x,N5,Td22x,N5)
            call dmurrv(N5,N5,Td22x,N5,n5,d2x,1,N5,AT)
            ax = AT(1)+AT(2)+AT(3)

            ATX = AX*AX
            IF (ATX .GE. ATN) THEN
               ATN = ATX
            ENDIF
         ENDDO

c      Critical values of standard Brownian motion 
c      taken from Shorack and Wellner (1986, p. 34)
c      b = 5.00   (10%)
c      b = 7.6343 (5%)   
c      b = 3.83   (1%)

            If (ATN/(AA*ah4) .ge.7.6343) then
               at1 = at1+1
            endif
            If (ATN/(AA*ah4) .ge.5.0) then
                at05 = at05+1
            endif
            If (ATN/(AA*ah4) .ge.3.83 ) then
                at01 = at01+1
            endif

            Write(6,*) 'Ling-Tong Test Statistic'
            write(6,*)  at1,  at05, at01

c         Box-Ljung testing
          ahh = 0.0
          do i=1,knk
             ahh = ahh+ry(i)*ry(i)/(aknk*h2(i))
          enddo

          ILL = 6
          do iL=1,ILL
             q(IL)   = 0.0
             arh(iL) = 0.0
          enddo

          do iL=1, ILL
             ail = float(iL)
             do i= iL+1, knk
                arh(iL) = arh(iL)+ry(i)*ry(i-iL)/(
     +                 (aknk-ail)*sqrt(h2(i)*h2(i-iL)))
             enddo
          enddo

         q(1) = aknk*(aknk+2)*arh(1)*arh(1)/(aknk-1)
         do iL=2,ILL
           q(IL) = q(IL-1)+aknk*(aknk+2)*arh(iL)*arh(iL)/(aknk-aiL)
         enddo

         If (q(6).ge.16.812) then  ! chi^{2}_{0.01} 6 df
            bl1 = bl1+1
         endif
         If (q(6).ge.12.592) then  ! chi^{2}_{0.05} 6 df
            bl05 = bl05+1
         endif
         If (q(6).ge.10.645) then  ! chi^{2}_{0.1} 6 df
             bl01 = bl01+1
         endif
         write(6,*)  'Box-Ljung Test Statisic'
         write(6,*)  Bl1, Bl05, Bl01
c         
c   Mak-Li test
         call test(ry,h2,hd,hdd,knk,dgg1,qm)

c  KL    Testing
c        call estimat1(y,knk,xx,g1,gg1,h2,ry,hdd,hh,HD,DD,DGG1)

c         do i=1, knk
c             eta(i) = ry(i)/SQRT(h2(i))
c             do ip=1,ir
c                w(ip, i) = hdd(ip,i)/h2(i)
c                write(6,200) w(ip,i)
c             enddo
c          enddo
c          call Testing(knk, eta, w, F,XF, AKN)

           IF (qm.GE.10.645) THEN ! chi^{2}_{0.1} 6 df
               RJECT10 = RJECT10+1
            ENDIF
           IF (qm.GE.12.592) THEN ! chi^{2}_{0.05} 6 df
               RJECT05 = RJECT05+1
            ENDIF
           IF (qm.GE.16.812) THEN ! chi^{2}_{0.01} 6 df
               RJECT01 = RJECT01+1
            ENDIF

            WRITE(6,*) RJECT01, RJECT05,RJECT10

   99  continue
         do i=1,n5
            write(6,100)  axx(i), sqrt(axx2(i)-axx(i)**2)
  100       format(' ',2F10.4)
         enddo

c  ******************************************

       stop
       end
c
       subroutine fcn(n5,xx,f)
       implicit real*8(a-h,o-z)
       parameter (nx=8000,ir=3)
       dimension xx(n5),g1(3),gg1(3,3),y(nx)
       dimension hdd(ir,nx),h2(nx),ry(nx)
       dimension dgg1(n5,n5),hd(1,nX),dd(1,NX)

       common  hh,g1,gg1,y,knk
       call estimat1(y,knk,xx,g1,gg1,h2,ry,hdd,hh,DGG1)
       f = hh
       return
       end
c
       subroutine grad(n5,xx,g)
       implicit real*8(a-h,o-z)
       parameter(nx=8000)
       dimension  xx(n5), g(n5),g1(3),gg1(3,3),y(nx)
       common  hh,g1,gg1,y,knk
       do i=1,n5
          g(i) = g1(i)
       enddo
       return
       end

       subroutine hess(n5,xx,gg,ldh)
       implicit real*8(a-h,o-z)
       parameter(nx=8000)
       dimension xx(n5),gg(ldh,n5),g1(3),gg1(3,3),y(nx)
       common  hh,g1,gg1,y,knk
       do i=1,n5
          do ii=1,n5
             gg(i,ii) = gg1(i,ii)
          enddo
       enddo
       return
       end

c      The following program  calculates gradien and information matrix 

       subroutine estimat1(y,knk,xx,g1,gg1,h2,ry,hdd,hh,dgg1)
       implicit real*8(a-h,o-z)
       parameter (nn=8000,ir=3,ip=1,n5=3)
       parameter (nx=8000)
       dimension y(nx),xx(n5),c(ir),ry(nn),h2(nn),dha(nn)
       dimension dh1(ir),dh2(ir,ir)
       dimension hdd(ir,nn),g1(n5),gg1(n5,n5),dgg1(n5,n5)

       aknk = float(knk)    
       do i=1,ir
          c(i) = xx(i)
       enddo

       do i=1,knk
          ry(i) = y(i)
       enddo
        
       h2(1) = 1
       do i=2,knk
          h2(1) = h2(1)+(ry(i)**2)/(aknk-2.0)
       enddo
       
       do i=2, knk
          h2(i) = 0.0
       enddo
       
       do i=2,knk 
          h2(i) = C(1)+c(3)*h2(i-1)+c(2)*(ry(i-1)**2)
       enddo
c
       hh=0
       do i=2,knk
          hh = hh+(log(h2(i))+ry(i)*ry(i)/h2(i))/(2*aknk)
       enddo
c        
       hdd(1,2) = 1/(1-c(3))
       hdd(2,2) = h2(2)/(1-c(3))
       hdd(3,2) = hdd(2,2)   

       do i=3,knk
          hdd(1,i) = 0.0
          hdd(2,i) = 0.0
          hdd(3,i) = 0.0
       enddo 

       do i=3,knk
          hdd(1,i) = 1+c(3)*hdd(1,i-1)
          hdd(2,i) = ry(i-1)**2+c(3)*hdd(2,i-1)
          hdd(3,i) = h2(i-1)+c(3)*hdd(3,i-1)
       enddo 

c  **********************************************************
c      The following is the information of delta       
c         
       do i=1,ir
          do ii=1,ir
             dh1(i)    = 0
             dh2(ii,i) = 0
          enddo
       enddo
       
       do i=3,knk
          do ii=1,ir
            dh1(ii) = dh1(ii)-(hdd(ii,i)*((ry(i)**2)/h2(i)-1))/(2*h2(i))
            do ii3=1,ir 
           dh2(ii,ii3) = dh2(ii,ii3)+(hdd(ii,i)*hdd(ii3,i)/(h2(i)**2))/2
            enddo
          enddo
       enddo
c  ***************************************************************
       aknk = float(knk)
       do i=1,n5
          g1(i) = dh1(i)/aknk
       enddo

       do i=1,n5
         do ii=1,n5
            gg1(i,ii) = -dh2(i,ii)/aknk
         enddo
       enddo

      call dlinrg(n5,gg1,n5,dgg1,n5)              
      return
      end

c    ************************************************************
       subroutine  Testing(knk,eta,w,F,XF,AKN)
       implicit real*8(a-h,o-z)
       parameter(nn=8000,ir=3)   
       dimension eta(nn),w(3,nn),ETA1(NN),WW(IR,IR),F(NN)
       dimension AK(ir, nn),WT(IR,IR),R(IR),XF(NN)
       
       do i=1,ir
          DO II=1,IR
            WW(I,II) = 0
          ENDDO
       ENDDO     

       DO J=1, KNK
          do i=1,ir
             DO II=1,IR
               WW(I,II) = WW(I,II)+W(I, J)*W(II, J)
             ENDDO
          ENDDO     
       ENDDO

       call dlinrg(IR,WW,IR,WT,IR)          

       AKN = 0
       DO I=2,4000             
          DO IP=1,IR    
             R(IP) = 0
          ENDDO           
          DO IP=1,IR    
             DO II=1,KNK                            
                AW = W(IP, II)*(0-F(I))
                IF (ETA(II) .LE. XF(I)) THEN
                   AW = W(IP, II)*(1-F(I))
                ENDIF
                R(IP) = R(IP)+ AW                        
             ENDDO
          ENDDO   
                 
          AA=dblinf(IR,IR,WT,IR, R,R)     
          IF (AKN .LE. AA) THEN
             AKN = AA
          ENDIF   
        ENDDO               
        WRITE(6,*) 'AKN'            
        WRITE(6,*) AKN            

        return
        end

c      ch1(3) = 7.815, ch(6) = 12.592, ch(12) = 21.026
c      ry is residual, h2=h^{2}_{t},
c      hd=dh/d theta_{1}, hdd=dh/d theta_{2}, dd=dg/d theta_{1}
c      dgg1 is the  information  matrix^{-1}, 
c      theta_{1}(ip) is parameters in mean part
c      theta_{2}(ir) is those c  in GARCH  part, n5=ir+ip
c      qm1--Li-Mak statistic, qm--Box-Ljung statistic
   
       subroutine test(ry,h2,hd,hdd,knk,dgg1,qm)
       implicit real*8(a-h,o-z)
       parameter(nx=8000,m=6,n5=3,ir=3)   
       dimension h2(nx),r(m),r1(m),ry(nx) 
       dimension dgg1(n5,n5),hd(1,nx),hdd(ir,nx)
       
       dimension x(n5,m),x1(n5,m),xt(m,n5),xt1(m,n5)
       dimension vv(m,m),v(n5,n5),vv1(m,m),vvt(m,m),vvt1(m,m)
       dimension vx(n5,m),vx1(n5,m)

       aknk = float(knk) 
       do i=1,m
          r(i) = 0
       enddo
       do i=1,m
          do ii=1,m
             vv(i,ii) = 0
          enddo
       enddo

       sc1=0.0
       do i=3, knk
          sc1 = sc1+((ry(i)**2)/h2(i))/aknk
       enddo
   
       do k=1,m
          do ii=1,ir
             x(ii,k) = 0
          enddo
        enddo  

       do k=1,m
          ak=float(k)
          do ii=1,ir
             do i=k+1,knk 
                x(ii,k) = x(ii,k)+( ry(i-k)*ry(i-k)/h2(i-k )-sc1 )
     +                  *hdd(ii,i)/(h2(i)*aknk)
              enddo   
           enddo
        enddo

         c0=0.0
         do i=1,knk
             c0=c0+(((ry(i)**2)/h2(i)-sc1)**2)/AKNK
         enddo

         do i=1,n5
             do ii=1,n5
                 v(i,ii)=0.0
             enddo
          enddo

          call dmrrrr(n5,n5,v,n5,n5,m,x,n5,n5,m,vx,n5)
          call dtrnrr(n5,m,x,n5,m,n5,xt,m)
          call dmrrrr(m,n5,xt,m,n5,m,vx,n5,m,m,vv,m)
          do i=1,m
               vv(i,i)=vv(i,i)+1.0
          enddo

          call dlinrg(m,vv,m,vvt,m)
  
          do i=1,m
             r(i)=0
          enddo

          do i=1,m
             ai=float(i)
             do k=i+1,knk
               r(i)=r(i)+((ry(k)**2)/h2(k)-sc1)*((ry(k-i)**2)/h2(k-i)
     +              -sc1)/(c0*(aknk-m))
              enddo
          enddo 

          qm=Aknk*dblinf(m,m,vvt,m,r,r)

          write(6,*) 'Li-Mak Test Statistic', qm
          return
          end

